from typing import List, Iterable, Set, Dict, Hashable, Tuple, FrozenSet, Union
Node = Hashable
DirectedEdge = Tuple[Node, Node]
UndirectedEdge = FrozenSet[Node]
BidirectedEdge = FrozenSet[Node]
NodeSet = Union[Hashable, Set[Hashable]]
